"""
Brewblox feature for the aiohttp client
"""

from typing import Optional

from aiohttp import ClientSession, web

from brewblox_service import features


class HTTPClient(features.ServiceFeature):
    def __init__(self, app: web.Application):
        super().__init__(app)
        self._session: Optional[ClientSession] = None

    @property
    def session(self) -> ClientSession:
        return self._session  # type: ignore

    async def startup(self, app: web.Application):
        await self.shutdown(app)
        self._session = await ClientSession(raise_for_status=True).__aenter__()

    async def shutdown(self, app: web.Application):
        if self._session:
            await self._session.__aexit__(None, None, None)
            self._session = None


def setup(app: web.Application):
    features.add(app, HTTPClient(app))


def get_client(app: web.Application) -> HTTPClient:
    return features.get(app, HTTPClient)


def session(app: web.Application) -> ClientSession:
    """Gets web.ClientSession for this app.

    Shortcut for `HTTPClient.session`.
    It requires setup() to have been called().
    """
    return get_client(app).session



"""
Abstract base class for a service that keeps repeating the same action.

This is useful for implementing broadcasting or polling services.

Example use:

    import asyncio
    from datetime import datetime
    from brewblox_service.repeater import RepeaterFeature, features, scheduler

    class Greeter(RepeaterFeature):
        async def prepare(self):
            print('Hello, I am starting now')

        async def run(self):
            await asyncio.sleep(5)
            print(datetime.now())

    def setup(app):
        scheduler.setup(app)
        features.add(app, Greeter(app))

"""

import asyncio
from abc import abstractmethod
from typing import Optional

from aiohttp import web

from brewblox_service import brewblox_logger, features, scheduler, strex

LOGGER = brewblox_logger(__name__)


class RepeaterCancelled(Exception):
    """
    This can be raised during either setup() or run() to permanently cancel execution.
    """


class RepeaterFeature(features.ServiceFeature):

    def __init__(self, app: web.Application):
        super().__init__(app)
        self._task: Optional[asyncio.Task] = None

    @property
    def active(self) -> bool:
        """
        Indicates whether the background task is currently running: not finished, not cancelled.
        """
        return bool(self._task and not self._task.done())

    async def startup(self, app: web.Application):
        await self.shutdown(app)
        await self.start()

    async def shutdown(self, app: web.Application):
        await self.end()

    async def __repeat(self):
        last_ok = True

        try:
            await self.prepare()

        except RepeaterCancelled:
            LOGGER.info(f'{self} cancelled during setup.')
            return

        except Exception as ex:
            LOGGER.error(f'{self} error during setup: {strex(ex)}')
            raise ex

        while True:
            try:
                await self.run()

                if not last_ok:
                    LOGGER.info(f'{self} resumed OK')
                    last_ok = True

            except RepeaterCancelled:
                LOGGER.info(f'{self} cancelled during runtime.')
                return

            except Exception as ex:
                # Only log the first error to prevent log spam
                if last_ok:
                    LOGGER.error(f'{self} error during runtime: {strex(ex)}')
                    last_ok = False

    async def start(self):
        """
        Initializes the repeater task.
        By default called during startup(), but implementations can choose to override this.
        Will cancel the previous task if called repeatedly.
        """
        await self.end()
        self._task = await scheduler.create(self.app, self.__repeat())

    async def end(self):
        """
        Ends the repeater task.
        Always called during shutdown(), but can be safely called earlier.
        """
        await scheduler.cancel(self.app, self._task)
        self._task = None

    @abstractmethod
    async def prepare(self):
        """
        One-time preparation.
        Any errors raised here will cause the repeater to abort.
        Raise RepeaterCancelled to abort without error logs.
        """

    @abstractmethod
    async def run(self):
        """
        This function will be called on repeat.
        It is advisable to implement rate limiting through the use of
        `await asyncio.sleep()`
        """



"""
Testing utility functions
"""

from aiohttp.client_exceptions import ContentTypeError


async def response(request, status=200):
    retv = await request
    if retv.status != status:
        print(retv)
        assert retv == status
    try:
        return await retv.json()
    except ContentTypeError:
        return await retv.text()



"""
Tests brewblox_service.scheduler
"""


import asyncio

import pytest

from brewblox_service import scheduler

TESTED = scheduler.__name__


@pytest.fixture
def app(app, mocker):
    mocker.patch(TESTED + '.CLEANUP_INTERVAL_S', 0.001)

    scheduler.setup(app)
    return app


async def test_create_cancel(app, client):

    async def do(ev):
        ev.set()
        return 'ok'

    ev = asyncio.Event()
    task = await scheduler.create(app, do(ev))
    await ev.wait()

    assert [
        await scheduler.cancel(app, task),
        await scheduler.cancel(app, task),
        await scheduler.cancel(app, task, False),
    ] == ['ok', 'ok', None]

    # Create and forget
    await scheduler.create(app, do(asyncio.Event()))

    # Cancelling None does not croak
    await scheduler.cancel(app, None)


async def test_cleanup(app, client):
    async def dummy():
        pass

    sched = scheduler.get_scheduler(app)
    start_count = len(sched._tasks)
    task = await scheduler.create(app, dummy())
    await asyncio.sleep(0.01)

    assert task.done()
    assert len(sched._tasks) == start_count



"""
Registers and gets features added to Aiohttp by brewblox services.
"""

from abc import ABC, abstractmethod
from enum import Enum, auto
from functools import wraps
from typing import Any, Hashable, Optional, Type, Union

from aiohttp import web

from brewblox_service import brewblox_logger

FEATURES_KEY = '#features'

LOGGER = brewblox_logger(__name__)


def add(app: web.Application,
        feature: Any,
        key: Optional[Union[Hashable, Type[Any]]] = None,
        exist_ok: bool = False
        ):
    """
    Adds a new feature to the app.

    Features can either be registered as the default feature for the class,
    or be given an explicit name.

    Args:
        app (web.Application):
            The current Aiohttp application.

        feature (Any):
            The new feature that should be registered.
            It is recommended, but not required to use a `ServiceFeature`.

        key (Hashable, optional):
            The key under which the feature should be registered.
            Defaults to `type(feature)`.

        exist_ok (bool):
            If truthy, this function will do nothing if a feature was already registered for `key`.
            Otherwise, an exception is raised.

    """
    if FEATURES_KEY not in app:
        app[FEATURES_KEY] = dict()

    key = key or type(feature)

    if key in app[FEATURES_KEY]:
        if exist_ok:
            return
        else:
            raise KeyError(f'Feature "{key}" already registered')

    app[FEATURES_KEY][key] = feature


def get(app: web.Application,
        feature_type: Optional[Type[Any]] = None,
        key: Optional[Hashable] = None
        ) -> Any:
    """
    Finds declared feature.
    Identification is done based on feature type and key.

    Args:
        app (web.Application):
            The current Aiohttp application.

        feature_type (Type[Any]):
            The Python type of the desired feature.
            If specified, it will be checked against the found feature.

        key (Hashable):
            A specific identifier for the desired feature.
            Defaults to `feature_type`

    Returns:
        Any: The feature found for the combination of `feature_type` and `key`
    """
    actual: Union[Hashable, Type[Any], None] = key or feature_type

    if actual is None:
        raise AssertionError('No feature identifier provided')

    try:
        found = app[FEATURES_KEY][actual]
    except KeyError:
        raise KeyError(f'No feature found for "{actual}"')

    if feature_type and not isinstance(found, feature_type):
        raise AssertionError(f'Found {found} did not match type "{feature_type}"')

    return found


class Startup(Enum):
    MANAGED = auto()
    MANUAL = auto()
    AUTODETECT = auto()


class ServiceFeature(ABC):
    """Base class for long-lived Aiohttp handler classes.

    For classes with async functionality,
    the (synchronous) `__init__()` and `__del__()` functions may not be sufficient.
    Aiohttp offers comparable init/deinit hooks, but inside the context of a running event loop.

    ServiceFeature registers the `self.startup(self, app)`, `self.before_shutdown(app)`,
     and `self.shutdown(self, app)` as lifecycle callbacks.
    They will be called by Aiohttp at the appropriate moment.
    By overriding these functions, subclasses can perform initialization/deinitialization that requires an event loop.

    Note: Aiohttp will not accept registration of new callbacks after it started running.
    Startup management can be adjusted by using the `startup` argument in `ServiceFeature.__init__()`

    Example class:

        import asyncio
        import random
        from aiohttp import web
        from brewblox_service import scheduler, service
        from brewblox_service.features import ServiceFeature

        class MyFeature(ServiceFeature):

            def __init__(self, app: web.Application):
                super().__init__(app)
                self._task: asyncio.Task = None

            async def startup(self, app: web.Application):
                # Schedule a long-running background task
                self._task = await scheduler.create(app, self._hello())

            async def before_shutdown(self, app: web.Application):
                print('Any minute now...')

            async def shutdown(self, app: web.Application):
                # Orderly cancel the background task
                await scheduler.cancel(app, self._task)

            async def _hello(self):
                while True:
                    await asyncio.sleep(5)
                    print(random.choice([
                        'Hellooo',
                        'Searching',
                        'Sentry mode activated',
                        'Is anyone there?',
                        'Could you come over here?',
                    ]))

    Example use:

        app = service.create_app(default_name='example')

        scheduler.setup(app)
        greeter = MyFeature(app)

        service.furnish(app)
        service.run(app)
        # greeter.startup(app) is called now

        # Press Ctrl+C to quit
        # greeter.before_shutdown(app) will be called
        # greeter.shutdown(app) will be called
    """

    def __hook(self, func, evt):
        @wraps(func)
        async def decorator(app):
            LOGGER.debug(f'--> {evt} {self}')
            retv = await func(app)
            LOGGER.debug(f'<-- {evt} {self}')
            return retv
        return decorator

    def __init__(self, app: web.Application, startup=Startup.AUTODETECT):
        """
        ServiceFeature constructor.

        Args:
            app (web.Application):
                The Aiohttp application with which the feature should be associated.

            startup (Startup):
                How feature lifecycle management should be handled. Default is AUTODETECT.
                    MANAGED:    Feature always registers lifecycle hooks.
                                This will raise an exception when creating
                                the feature while the application is running.

                    MANUAL:     Feature will not register lifecycle hooks.
                                startup() and shutdown() must be called manually.

                    AUTODETECT: Feature will register lifecycle hooks only if app is not running.
                                Behaves like MANAGED before application start,
                                and like MANUAL after application start.

        """
        self.__active_app: web.Application = app

        if any([
            startup == Startup.MANAGED,
            startup == Startup.AUTODETECT and not app.frozen
        ]):
            app.on_startup.append(self.__hook(self.startup, 'startup'))
            app.on_shutdown.append(self.__hook(self.before_shutdown, 'before_shutdown'))
            app.on_cleanup.append(self.__hook(self.shutdown, 'shutdown'))

    def __str__(self):
        return f'<{type(self).__name__}>'

    @property
    def app(self) -> web.Application:
        """Currently active `web.Application`

        Returns:
            web.Application: The current app.
        """
        return self.__active_app

    @abstractmethod
    async def startup(self, app: web.Application):
        """Lifecycle hook for initializing the feature in an async context.

        Subclasses are expected to override this function.

        Depending on the `startup` argument in `__init__()`,
        `startup()` will be called when Aiohttp starts running.

        Args:
            app (web.Application):
                Current Aiohttp application.
        """

    async def before_shutdown(self, app: web.Application):
        """Lifecycle hook for preparing to shut down the feature.

        Subclasses may override this function, but it is not mandatory.

        Depending on the `startup` argument in `__init__()`,
        `before_shutdown()` will be called when Aiohttp is closing.

        Args:
            app (web.Application):
                Current Aiohttp application.
        """

    @abstractmethod
    async def shutdown(self, app: web.Application):
        """Lifecycle hook for shutting down the feature before the event loop is closed.

        Subclasses are expected to override this function.

        Depending on the `startup` argument in `__init__()`,
        `shutdown()` will be called when Aiohttp is closing.

        Args:
            app (web.Application):
                Current Aiohttp application.
        """



"""
Tests brewblox_service.testing
"""

import pytest

from brewblox_service import service, testing


@pytest.fixture
async def app(app, mocker):
    service.furnish(app)
    return app


async def test_response(app, client):
    await testing.response(client.get('/test_app/_service/status'))
    with pytest.raises(AssertionError):
        await testing.response(client.get('/test_app/_service/status'), 400)



import asyncio
from typing import Any, Optional, Tuple

from aiohttp import client_exceptions, web

from brewblox_service import brewblox_logger, features, http

LOGGER = brewblox_logger(__name__)

DB_RETRY_INTERVAL_S = 1
COUCH_URL = 'http://datastore:5984'


class CouchDBClient(features.ServiceFeature):

    def __str__(self):
        return f'<{type(self).__name__} for {COUCH_URL}>'

    async def startup(self, app: web.Application):
        pass

    async def shutdown(self, app: web.Application):
        pass

    async def check_remote(self):
        session = http.session(self.app)
        num_attempts = 0
        while True:
            try:
                await session.head(COUCH_URL, raise_for_status=True)
                break
            except Exception:
                num_attempts += 1
                if num_attempts % 10 == 0:
                    LOGGER.info(f'{self} Waiting for datastore...')
                await asyncio.sleep(DB_RETRY_INTERVAL_S)

    async def read(self,
                   database: str,
                   document: str,
                   default_data: Any,
                   ) -> Tuple[str, Any]:
        db_url = f'{COUCH_URL}/{database}'
        document_url = f'{db_url}/{document}'
        session = http.session(self.app)

        async def ensure_database():
            try:
                await session.put(db_url)
                LOGGER.info(f'{self} New database created ({database})')

            except client_exceptions.ClientResponseError as ex:
                if ex.status != 412:  # Already exists
                    raise ex

        async def create_document() -> Optional[Tuple[str, Any]]:
            try:
                resp = await session.put(document_url, json={'data': default_data})
                resp_content = await resp.json()

                rev = resp_content['rev']
                data = default_data
                LOGGER.info(f'{self} New document created ({document})')
                return rev, data

            except client_exceptions.ClientResponseError as ex:
                if ex.status == 409:  # Conflict: already exists
                    return None
                else:
                    raise ex

        async def read_document() -> Optional[Tuple[str, Any]]:
            try:
                resp = await session.get(document_url)
                resp_content = await resp.json()

                rev = resp_content['_rev']
                data = resp_content['data']
                LOGGER.info(f'{self} Existing document found ({document})')
                return rev, data

            except client_exceptions.ClientResponseError as ex:
                if ex.status == 404:
                    return None
                else:
                    raise ex

        try:
            await self.check_remote()
            await ensure_database()
            read_result, create_result = await asyncio.gather(read_document(), create_document())
            (rev, data) = read_result or create_result or (None, None)
            if rev is None:
                raise ValueError('Data was neither read nor created')
            return rev, data

        except Exception as ex:
            LOGGER.error(f'{self} {type(ex).__name__}({ex})')
            raise ex

    async def write(self, database: str, document: str, rev: str, data: Any) -> str:
        resp = await http.session(self.app).put(
            url=f'{COUCH_URL}/{database}/{document}',
            json={'data': data},
            params=[('rev', rev)]
        )
        resp_content = await resp.json()
        return resp_content['rev']


def setup(app: web.Application):
    """Enables CouchDB interaction.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    features.add(app, CouchDBClient(app))


def get_client(app: web.Application) -> CouchDBClient:
    """Gets registered CouchDBClient.
    Requires setup() to have been called first.

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    return features.get(app, CouchDBClient)


async def check_remote(app: web.Application):
    """Waits for the remote datastore to be available

    Shortcut for `CouchDBClient.check_remote()`

    Args:
        app (web.Application):
            The Aiohttp Application object.
    """
    await get_client(app).check_remote()


async def read(app: web.Application,
               database: str,
               document: str,
               default_data: Any
               ) -> Tuple[str, Any]:
    """Fetches data from CouchDB document.

    If the database does not exist, it will be created.
    If the document does not exist, it will be created, and filled with `default_data`.

    Shortcut for `CouchDBClient.read()`
    It requires setup() to have been called.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        database (str):
            The CouchDB database name.

        document (str):
            The document name in the database.

        default_data (Any):
            If the document does not exist, it will be initialized with this.

    Returns:
        (revision ID, data):
            The revision ID is required for later writing data to the document.
            Data is document content, and equals `default_data` if the document was created.
    """
    return await get_client(app).read(database,
                                      document,
                                      default_data)


async def write(app: web.Application,
                database: str,
                document: str,
                rev: str,
                data: Any
                ) -> str:
    """Writes data to an existing CouchDB document.

    Shortcut for `CouchDBClient.write()`
    It requires setup() to have been called.

    Args:
        app (web.Application):
            The Aiohttp Application object.

        database (str):
            The CouchDB database name.

        document (str):
            The document name in the database.

        rev (str):
            The document revision ID.
            It must match that of the last update to the document.

        data (any):
            Document content.

    Returns:
        str:
            The new revision ID.
    """
    return await get_client(app).write(database,
                                       document,
                                       rev,
                                       data)



"""
Tests functionality offered by brewblox_service.events
"""

import asyncio
import json
from datetime import timedelta
from unittest.mock import AsyncMock, Mock, call

import aioamqp
import pytest

from brewblox_service import events, scheduler
from brewblox_service.testing import response

TESTED = events.__name__


@pytest.fixture
def m_channel():
    m = Mock()
    m.exchange_declare = AsyncMock()
    m.queue_declare = AsyncMock(return_value={'queue': 'queue_name'})
    m.queue_bind = AsyncMock()
    m.basic_consume = AsyncMock()
    m.basic_publish = AsyncMock()
    m.basic_client_ack = AsyncMock()
    return m


@pytest.fixture
def m_protocol(m_channel):
    m = Mock()
    m.channel = AsyncMock(return_value=m_channel)
    m.ensure_open = AsyncMock()
    m.close = AsyncMock()
    return m


@pytest.fixture
def m_transport():
    return Mock()


@pytest.fixture
def mocked_connect(mocker, m_protocol, m_transport):
    # connect()
    conn_mock_func = AsyncMock(spec=aioamqp.connect)
    conn_mock_func.return_value = (m_transport, m_protocol)

    mocker.patch.object(events.aioamqp, 'connect', conn_mock_func)

    return conn_mock_func


@pytest.fixture
async def app(app, mocker, loop, mocked_connect):
    """App with events enabled"""
    scheduler.setup(app)
    events.setup(app)
    mocker.patch(TESTED + '.RECONNECT_INTERVAL', timedelta(microseconds=100))
    return app


async def test_setup(app, client):
    assert events.get_listener(app)
    assert events.get_publisher(app)


async def test_subscribe_callbacks(app, client, mocker, m_channel):
    cb = Mock()
    content = dict(key='var')

    body = json.dumps(content)
    envelope_mock = Mock()
    envelope_mock.routing_key = 'message_key'
    properties_mock = Mock()

    sub = events.get_listener(app).subscribe(
        'brewblox', 'router', on_message=cb)

    await sub._relay(m_channel, body, envelope_mock, properties_mock)
    assert m_channel.basic_client_ack.call_count == 1
    cb.assert_called_once_with(sub, 'message_key', content)

    cb2 = Mock()
    sub.on_message = cb2

    await sub._relay(m_channel, body, envelope_mock, properties_mock)
    cb2.assert_called_once_with(sub, 'message_key', content)
    assert m_channel.basic_client_ack.call_count == 2
    assert cb.call_count == 1

    # Shouldn't blow up
    cb2.side_effect = ValueError
    await sub._relay(m_channel, body, envelope_mock, properties_mock)
    assert m_channel.basic_client_ack.call_count == 3

    # Should be tolerated
    sub.on_message = None
    await sub._relay(m_channel, body, envelope_mock, properties_mock)
    assert m_channel.basic_client_ack.call_count == 4


async def test_offline_listener(app, mocker):
    # We're not importing the client fixture, so the app is not running
    # Expected behaviour is for the listener to add functions to the app hooks
    listener = events.EventListener(app)

    # Subscriptions will not be declared yet - they're waiting for startup
    listener.subscribe('exchange', 'routing')
    assert listener._pending.qsize() == 1
    assert not listener.active


async def test_deferred_subscription(app, client):
    # Tests whether subscriptions made before startup are correctly initialized
    listener = events.EventListener(app)
    assert not listener.active
    listener.subscribe('exchange', 'route')
    await listener.prepare()
    assert listener._has_pending.is_set()


async def test_online_listener(app, client, mocker):
    listener = events.get_listener(app)
    events.subscribe(app, 'exchange', 'routing')

    pending_subs = listener._pending.qsize()
    events.subscribe(app, 'exchange', 'routing')
    # We haven't yielded, so the subscription can't have been processed yet
    assert listener._pending.qsize() == pending_subs + 1

    # Safe for repeated calls
    await listener.shutdown(app)
    await listener.shutdown(app)


async def test_offline_publisher(app):
    publisher = events.EventPublisher(app)
    await publisher.startup(app)
    await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_online_publisher(app, client, mocker):
    publisher = events.EventPublisher(app)
    await publisher.startup(app)

    await publisher.publish('exchange', 'key', message=dict(key='val'))
    await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_publish_endpoint(app, client, mocker, m_channel):
    # standard ok
    await response(client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='first',
        message=dict(key='val')
    )))

    # string messages supported
    await response(client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='second',
        message='message'
    )))

    mocker.patch.object(events.get_publisher(app),
                        '_ensure_channel',
                        AsyncMock(side_effect=RuntimeError))

    await response(client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='first',
        message=dict(key='val')
    )), 500)

    msg1 = json.dumps({'key': 'val'}).encode()
    msg2 = json.dumps('message').encode()

    m_channel.basic_publish.assert_has_calls([
        call(payload=msg1, exchange_name='exchange', routing_key='first'),
        call(payload=msg2, exchange_name='exchange', routing_key='second'),
    ])


async def test_subscribe_endpoint(app, client, m_channel):
    await response(client.post('/_debug/subscribe', json=dict(
        exchange='exchange',
        routing='routing.key'
    )))

    await asyncio.sleep(0.01)

    m_channel.queue_bind.assert_awaited_once_with(
        queue_name='queue_name',
        exchange_name='exchange',
        routing_key='routing.key'
    )


async def test_listener_exceptions(app, client, m_protocol, m_channel, m_transport, mocked_connect):
    listener = events.get_listener(app)

    m_channel.basic_consume.side_effect = ConnectionRefusedError
    listener.subscribe('exchange', 'routing')

    # Let listener attempt to declare the new subscription a few times
    # In real scenarios we'd expect ensure_open() to also start raising exceptions
    await asyncio.sleep(0.01)
    m_protocol.ensure_open.side_effect = aioamqp.AmqpClosedConnection
    await asyncio.sleep(0.01)

    # Retrieved subscription, errored on basic_consume(), and put subscription back
    assert m_channel.basic_consume.call_count >= 1
    assert listener._pending.qsize() == 1
    assert not listener._task.done()

    # Error recovery
    m_protocol.ensure_open.side_effect = RuntimeError
    await asyncio.sleep(0.01)
    assert not listener._task.done()

    # Should be closed every time it was opened
    await listener.shutdown(app)
    assert m_protocol.close.call_count == mocked_connect.call_count
    assert m_transport.close.call_count == mocked_connect.call_count


async def test_listener_periodic_check(mocker, app, client, m_protocol):
    # No subscriptions were made - not listening
    await asyncio.sleep(0.1)
    assert m_protocol.ensure_open.call_count == 0

    listener = events.get_listener(app)

    # Should be listening now, an periodically calling ensure_open()
    listener.subscribe('exchange', 'routing')
    await asyncio.sleep(0.1)
    assert m_protocol.ensure_open.call_count > 0


async def test_publisher_exceptions(app, client, m_protocol):
    m_protocol.ensure_open.side_effect = aioamqp.AmqpClosedConnection

    with pytest.raises(aioamqp.AmqpClosedConnection):
        await events.get_publisher(app).publish('exchange', 'routing', 'message')
